/*
*	SmallBasic LIBRARY - GRAPHICS
*
*	2000-05-27, Nicholas Christopoulos
*
*	This program is distributed under the terms of the GPL v2.0 or later
*	Download the GNU Public License (GPL) from www.gnu.org
*/

#include "sys.h"
#include "str.h"
#include "kw.h"
#include "panic.h"
#include "var.h"
#include "blib.h"
#include "pproc.h"

// graphics - rel. coord.
static int  gra_x;
static int  gra_y;

//
void	graph_reset()
{
	gra_x = gra_y = 0;
	dev_cls();
}

//
//	PSET [STEP] x, y [COLOR color]
//
void	cmd_pset()
{
	int		x1, y1;
	int		color = dev_fgcolor, step1 = 0;

	/* [STEP] x, y */
	if	( code_peek() == kwSTEP )	{
		code_skipnext();
		step1 = 1;
		}

	x1 = par_getint();	if ( prog_error )	return;
	par_getcomma();		if ( prog_error )	return;
	y1 = par_getint();	if ( prog_error )	return;
	if	( step1 )
		(x1 += gra_x, y1 += gra_y);

	if	( code_peek() == kwCOLOR )	{
		code_skipnext();
		color = par_getint();	if ( prog_error )	return;
		}

	gra_x = x1;
	gra_y = y1;

	if	( color != dev_fgcolor )	{
		int		prev_color = dev_fgcolor;

		dev_setcolor(color);
		dev_setpixel(x1, y1);
		dev_setcolor(prev_color);
		}
	else
		dev_setpixel(x1, y1);
}

//
//	LINE [STEP] x, y [{,|STEP} x2, y2] [COLOR color]
//
void	cmd_line()
{
	int		x1, y1,	x2, y2;
	int		step1 = 0, step2 = 0;
	int		color = dev_fgcolor;
	byte	code;

	/* [STEP] x, y */
	code = code_peek();
	if	( code == kwSTEP )	{
		code_skipnext();
		step1 = 1;
		}

	x1 = par_getint();	if ( prog_error )	return;
	par_getcomma();		if ( prog_error )	return;
	y1 = par_getint();	if ( prog_error )	return;
	if	( step1 )
		( x1 += gra_x, y1 += gra_y );

	code = code_peek();
	if	( code == kwTYPE_SEP || code == kwSTEP )	{
		/* x, y [STEP] ? */
		if	( code == kwSTEP )	{
			code_skipnext();
			step2 = 1;
			}
		else
			par_getcomma();		if ( prog_error )	return;

		/* [STEP] x, y [,|STEP] x2, y2 */
		x2 = par_getint();	if ( prog_error )	return;
		par_getcomma();		if ( prog_error )	return;
		y2 = par_getint();	if ( prog_error )	return;
		if	( step2 )
			( x2 += x1, y2 += y1 );
		}
	else	{
		x2 = x1;
		y2 = y1;
		x1 = gra_x;
		y1 = gra_y;
		}

	code = code_peek();
	if	( code == kwCOLOR )	{
		code_skipnext();
		color = par_getint();	if ( prog_error )	return;
		}

	/* draw */
	gra_x = x2;
	gra_y = y2;

	if	( color != dev_fgcolor )	{
		int		prev_color = dev_fgcolor;

		dev_setcolor(color);
		dev_line(x1, y1, x2, y2);
		dev_setcolor(prev_color);
		}
	else
		dev_line(x1, y1, x2, y2);
}

//
//	RECT [STEP] x, y [{,|STEP} x2, y2] [COLOR color] [FILLED]
//
void	cmd_rect()
{
	int		x1, y1,	x2, y2, fill = 0;
	int		step1 = 0, step2 = 0;
	int		color = dev_fgcolor;
	byte	code;

	/* [STEP] x, y */
	code = code_peek();
	if	( code == kwSTEP )	{
		code_skipnext();
		step1 = 1;
		}

	x1 = par_getint();	if ( prog_error )	return;
	par_getcomma();		if ( prog_error )	return;
	y1 = par_getint();	if ( prog_error )	return;
	if	( step1 )
		( x1 += gra_x, y1 += gra_y );

	code = code_peek();
	if	( code == kwTYPE_SEP || code == kwSTEP )	{
		/* x, y [STEP] ? */
		if	( code == kwSTEP )	{
			code_skipnext();
			step2 = 1;
			}
		else	{
			par_getcomma();		
			if ( prog_error )	return;
			}

		/* [STEP] x, y [,|STEP] x2, y2 */
		x2 = par_getint();	if ( prog_error )	return;
		par_getcomma();		if ( prog_error )	return;
		y2 = par_getint();	if ( prog_error )	return;
		if	( step2 )
			( x2 += x1, y2 += y1 );
		}
	else	{
		x2 = x1;
		y2 = y1;
		x1 = gra_x;
		y1 = gra_y;
		}

	if	( code_peek() == kwCOLOR )	{
		code_skipnext();
		color = par_getint();	if ( prog_error )	return;
		}
	if	( code_peek() == kwFILLED )	{
		code_skipnext();
		fill = 1;
		}

	/* draw */
	gra_x = x2;
	gra_y = y2;

	if	( color != dev_fgcolor )	{
		int		prev_color = dev_fgcolor;

		dev_setcolor(color);
		dev_rect(x1, y1, x2, y2, fill);
		dev_setcolor(prev_color);
		}
	else
		dev_rect(x1, y1, x2, y2, fill);
}

//
//	DRAWPOLY v() [, xorg, yorg [, scale] [COLOR color] // [FILLED]
//
//	0,2,4... x 
//  1,3,5... y
//
void	cmd_drawpoly()
{
	var_t	*var_p, *elem_p;
	int		i, count;
	int		*pts, xorg = 0, yorg = 0;
	int		prev_color = dev_fgcolor;
	int		color = dev_fgcolor;
	int		filled = 0, scalef = 0;
	double	scale = 1.0;

	// array
	if	( !code_isnullarray() )	{
		err_typemismatch();
		return;
		}
	code_skipnext();
	var_p = tvar[code_getnext16()];
	count = var_p->size;

	// x,y origin
	if	( code_peek() == kwTYPE_SEP )	{
		par_getcomma();			if	( prog_error )	return;
		xorg = par_getint();	if	( prog_error )	return;
		par_getcomma();			if	( prog_error )	return;
		yorg = par_getint();	if	( prog_error )	return;

		// scale factor
		if	( code_peek() == kwTYPE_SEP )	{
			par_getcomma();			if	( prog_error )	return;
			scale = par_getnum();	if	( prog_error )	return;
			scalef ++;
			}
		}

	// color
	if	( code_peek() == kwCOLOR )	{
		code_skipnext();
		color = par_getint();	if ( prog_error )	return;
		}

	// filled
	if	( code_peek() == kwFILLED )	{
		code_skipnext();
		filled ++;
		}

	// get array's values
	pts = (int *) tmp_alloc(sizeof(int) * count);
	if	( scalef == 0 )	{
		for ( i = 0; i < count; i ++ )	{
			elem_p = v_getelemptr(var_p, i);
			if	( prog_error )	{
				tmp_free(pts);
				return;
				}

			switch ( elem_p->type )	{
			case	V_INT:
				pts[i] = elem_p->i;
				break;
			case	V_NUM:
				pts[i] = elem_p->n;
				break;
			case	V_STR:
				pts[i] = v_getval(elem_p);
				break;
			default:
				err_typemismatch();
				tmp_free(pts);
				return;
				}

			pts[i] += ( ((i % 2) == 0) ? xorg : yorg);
			}
		}
	else	{
		for ( i = 0; i < count; i ++ )	{
			elem_p = v_getelemptr(var_p, i);
			if	( prog_error )	{
				tmp_free(pts);
				return;
				}

			switch ( elem_p->type )	{
			case	V_INT:
				pts[i] = ((double) elem_p->i) * scale;
				break;
			case	V_NUM:
				pts[i] = elem_p->n * scale;
				break;
			case	V_STR:
				pts[i] = v_getval(elem_p) * scale;
				break;
			default:
				err_typemismatch();
				tmp_free(pts);
				return;
				}

			pts[i] += ( ((i % 2) == 0) ? xorg : yorg);
			}
		}

	// ready
	if	( color != dev_fgcolor )
		dev_setcolor(color);

	count = count >> 1;
	for ( i = 1; i < count; i ++ )	
		dev_line(pts[(i-1)*2], pts[(i-1)*2+1], pts[i*2], pts[i*2+1]);

	// cleanup
	tmp_free(pts);

	if ( color != prev_color )
		dev_setcolor(prev_color);
}

//
//	CIRCLE [STEP] x, y, r [, aspect] [COLOR color] [FILLED]
//
void	cmd_circle()
{
	int		color = dev_fgcolor, fill = 0, step1 = 0;
	int		xc, yc, r;
	double	aspect = 1.0;
	byte	code;

	/* [STEP] x, y */
	code = code_peek();
	if	( code == kwSTEP )	{
		code_skipnext();
		step1 = 1;
		}

	// xc
	xc = par_getint();	if ( prog_error )	return;
	par_getcomma();		if ( prog_error )	return;
	// yc
	yc = par_getint();	if ( prog_error )	return;
	par_getcomma();		if ( prog_error )	return;
	// r
	r = par_getint();	if ( prog_error )	return;
	if	( step1 )
		( xc += gra_x, yc += gra_y );

	//  aspect
	if	( code_peek() == kwTYPE_SEP )	{
		par_getcomma();				if ( prog_error )	return;
		aspect = par_getnum();		if ( prog_error )	return;
		}

	// COLOR
	if	( code_peek() == kwCOLOR )	{
		code_skipnext();
		color = par_getint();	if ( prog_error )	return;
		}

	// FILLED
	if	( code_peek() == kwFILLED )	{
		code_skipnext();
		fill = 1;
		}

	if	( color != dev_fgcolor )	{
		int		prev_color = dev_fgcolor;

		dev_setcolor(color);
		dev_ellipse(xc, yc, r, r, aspect, fill);
		dev_setcolor(prev_color);
		}
	else	
		dev_ellipse(xc, yc, r, r, aspect, fill);
}

//
//	ARC [STEP] x, y, r, start, end [, aspect] [COLOR color]
//
void	cmd_arc()
{
	int		color = dev_fgcolor, step1 = 0;
	int		xc, yc, r;
	double	as, ae, aspect = 1.0;
	byte	code;

	/* [STEP] x, y */
	code = code_peek();
	if	( code == kwSTEP )	{
		code_skipnext();
		step1 = 1;
		}

	// x
	xc = par_getint();	if ( prog_error )	return;

	// y
	par_getcomma();		if ( prog_error )	return;
	yc = par_getint();	if ( prog_error )	return;

	// r
	par_getcomma();		if ( prog_error )	return;
	r = par_getint();	if ( prog_error )	return;

	// a.st.
	par_getcomma();		if ( prog_error )	return;
	as = par_getnum();	if ( prog_error )	return;

	// a.end
	par_getcomma();		if ( prog_error )	return;
	ae = par_getnum();	if ( prog_error )	return;

	if	( step1 )
		( xc += gra_x, yc += gra_y );

	// aspect
	if	( code_peek() == kwTYPE_SEP )	{
		par_getcomma();			if ( prog_error )	return;
		aspect = par_getnum();	if ( prog_error )	return;
		}

	if	( code_peek() == kwCOLOR )	{
		code_skipnext();
		color = par_getint();	if ( prog_error )	return;
		}

	if	( color != dev_fgcolor )	{
		int		prev_color = dev_fgcolor;

		dev_setcolor(color);
		dev_arc(xc, yc, r, as, ae, aspect);
		dev_setcolor(prev_color);
		}
	else	
		dev_arc(xc, yc, r, as, ae, aspect);
}

//
char	*draw_getval(const char *src, int *c)
{
	char	*p = (char *) src;
	char	*dst, buf[64];

	dst = buf;
	p ++;
	*c = 0;
	if	( *p == '-' )	{
		*dst = '-';
		dst ++;
		p ++;
		}

	while ( is_digit(*p) )	
		*dst ++ = *p ++;
	*dst = '\0';

	*c = xstrtol(buf);
	return p;
}

//
//	DRAW "commands"
//
void	cmd_draw()
{
	register int draw=1, update=1;
	int		x, y, r;
	int		prev_color = dev_fgcolor;
	char	*p;
	var_t	var;

	par_getstr(&var);	if ( prog_error )	return;
	p = var.ptr;
	while ( *p )	{

		// commands prefix
		while ( strchr("BbNn", *p) )	{
			if	( *p == 'B' || *p == 'b' )	{	// do not draw
				draw = 0;
				p ++;
				}
			else
				draw = 1;

			if	( *p == 'N' || *p == 'n' )	{	// do not update the position
				update = 0;
				p ++;
				}
			else
				update = 1;
			}

		// commands
		switch ( *p )	{
		case 'U': case 'u':	// up
			p = draw_getval(p, &y);
			if	( draw )
				dev_line(gra_x, gra_y, gra_x, gra_y - y);
			if	( update )
				gra_y -= y;
			continue;
		case 'D': case 'd':	// down
			p = draw_getval(p, &y);
			if	( draw )
				dev_line(gra_x, gra_y, gra_x, gra_y + y);
			if	( update )
				gra_y += y;
			continue;
		case 'L': case 'l':	// left
			p = draw_getval(p, &x);
			if	( draw )
				dev_line(gra_x, gra_y, gra_x - x, gra_y);
			if	( update )
				gra_x -= x;
			continue;
		case 'R': case 'r':	// right
			p = draw_getval(p, &x);
			if	( draw )
				dev_line(gra_x, gra_y, gra_x + x, gra_y);
			if	( update )
				gra_x += x;
			continue;
		case 'E': case 'e':	// up & right
			p = draw_getval(p, &x);
			if	( draw )
				dev_line(gra_x, gra_y, gra_x + x, gra_y - x);
			if	( update )	{
				gra_x += x;
				gra_y -= x;
				}
			continue;
		case 'F': case 'f':	// down & right
			p = draw_getval(p, &x);
			if	( draw )
				dev_line(gra_x, gra_y, gra_x + x, gra_y + x);
			if	( update )	{
				gra_x += x;
				gra_y += x;
				}
			continue;
		case 'G': case 'g':	// down & left
			p = draw_getval(p, &x);
			if	( draw )
				dev_line(gra_x, gra_y, gra_x - x, gra_y + x);
			if	( update )	{
				gra_x -= x;
				gra_y += x;
				}
			continue;
		case 'H': case 'h':	// up & left
			p = draw_getval(p, &x);
			if	( draw )
				dev_line(gra_x, gra_y, gra_x - x, gra_y - x);
			if	( update )	{
				gra_x -= x;
				gra_y -= x;
				}
			continue;
		case 'M': case 'm':	// move to x, y
			if	( *(p+1) == '-' )	{	// relative
				r = -1;
				p ++;
				}
			if	( *(p+1) == '+' )	{	// relative
				r = 1;
				p ++;
				}
			else
				r = 0;					// absolute

			p = draw_getval(p, &x);
			if	( *p != ',' )	{
				rt_raise("DRAW: MISSING ,");
				v_free(&var);
				return;
				}
			else
				p ++;
			p = draw_getval(p, &y);

			if	( r )	{
				if	( draw )
					dev_line(gra_x, gra_y, gra_x + x * r, gra_y + y * r);
				if	( update )	{
					gra_x += x * r;
					gra_y += x * r;
					}
				}
			else	{
				if	( draw )
					dev_line(gra_x, gra_y, x, y);
				if	( update )	{
					gra_x = x;
					gra_y = y;
					}
				}
			continue;
		case 'C': case 'c':	// color
			p = draw_getval(p, &x);
			dev_setcolor(x);
			continue;
		default:
			rt_raise("DRAW: '%c' UNSUPPORTED", *p);
			v_free(&var);
			return;
			}
		p ++;
		}

	dev_setcolor(prev_color);
	v_free(&var);
}

//
//	CHART chart-type, v() [, mark-type [, x1, y1, x2, y2]]
//
//  chart-type
//		1 = line-chart
//		2 = bar-chart
//
//	mark-type (bit-mask)
//		0 = none
//		1 = labels
//		2 = ruler
//
void	cmd_chart_fstr(double v, char *buf)
{
	if	( fabs(v) >= 10000000000.0 )	{
		ftostr(v / 1000000000.0, buf);
		if	( buf[3] == '.' )
			buf[3] = '\0';
		else
			buf[4] = '\0';
		strcat(buf, "G");
		}
	else if	( fabs(v) >= 10000000.0 )	{
		ftostr(v / 1000000.0, buf);
		if	( buf[3] == '.' )
			buf[3] = '\0';
		else
			buf[4] = '\0';
		strcat(buf, "M");
		}
	else if	( fabs(v) >= 10000.0 )	{
		ftostr(v / 1000.0, buf);
		if	( buf[3] == '.' )
			buf[3] = '\0';
		else
			buf[4] = '\0';
		strcat(buf, "K");
		}
	else	{
		ftostr(v, buf);
		buf[5] = '\0';
		}
}

void	cmd_chart()
{
	var_t	*var_p, *elem_p;
	int		i, count;
	double	*vals;
	double	vmin, vmax, lx, ly;
	int		dx, dy, color = 0;
	int		*pts;
	char	buf[32];
	int		prev_fgcolor = dev_fgcolor;
	int		prev_bgcolor = dev_bgcolor;

	// parameters
	int		chart = 0, marks = 0;
	int		x1 = 0, y1 = 0, x2 = os_graf_mx, y2 = os_graf_my;
	int		rx1 = 0, ry2;

	// chart type
	chart = par_getint();
	if	( prog_error )	return;

	par_getcomma();
	if	( prog_error )	return;

	// array
	if	( !code_isnullarray() )	{
		err_typemismatch();
		return;
		}
	code_skipnext();
	var_p = tvar[code_getnext16()];
	count = var_p->size;

	// optional labels-flag
	if	( code_peek() == kwTYPE_SEP )	{
		par_getcomma();				if	( prog_error )		return;
		marks = par_getint();		if	( prog_error )		return;

		// optional x1,y1,x2,y2
		if	( code_peek() == kwTYPE_SEP )	{
			par_getcomma();			if	( prog_error )		return;
			rx1 = x1 = par_getint();		if	( prog_error )		return;
			par_getcomma();			if	( prog_error )		return;
			y1 = par_getint();		if	( prog_error )		return;
			par_getcomma();			if	( prog_error )		return;
			x2 = par_getint();		if	( prog_error )		return;
			par_getcomma();			if	( prog_error )		return;
			ry2 = y2 = par_getint();		if	( prog_error )		return;
			}
		}

	// get array's values
	vals = (double *) tmp_alloc(sizeof(double) * count);
	for ( i = 0; i < count; i ++ )	{
		elem_p = v_getelemptr(var_p, i);
		if	( prog_error )	{
			tmp_free(vals);
			return;
			}

		switch ( elem_p->type )	{
		case	V_INT:
			vals[i] = elem_p->i;
			break;
		case	V_NUM:
			vals[i] = elem_p->n;
			break;
		case	V_STR:
			vals[i] = v_getval(elem_p);
			break;
		default:
			err_typemismatch();
			tmp_free(vals);
			return;
			}
		}

	// ready
	dev_settextcolor(0, 15);
	pts = (int *) tmp_alloc(sizeof(int) * count * 2);

	if	( marks & 0x2 )	{	// ruler
		x1 += dev_textwidth("00000") + 1;
		y2 -= (dev_textheight("0") + 1);
		}

	if	( marks & 0x1 )	{
		if	( chart == 1 )	{	// line
			x1 += 2;	x2 -= 2;
			y1 += 2;	y2 -= 2;
			}
		}

	dx = (x2-x1);
	dy = (y2-y1);

	// limits
	vmin = vmax = vals[0];
	for ( i = 1; i < count; i ++ )	{
		if	( vmin > vals[i] )
			vmin = vals[i];
		if	( vmax < vals[i] )
			vmax = vals[i];
		}

	if	( chart == 1 )	// line-chart
		lx = ((double) dx) / (double) (count-1);
	else
		lx = ((double) dx) / (double) count;

	ly = ((double) dy) / (vmax-vmin);

	// calc points
	for ( i = 0; i < count; i ++ )	{
		pts[i*2]   = x1 + i*lx;
		pts[i*2+1] = y1 + (dy - ((vals[i]-vmin)*ly));
		}

	// draw ruler
	if	( marks & 0x2 )	{
		int		fh, fw, n, y, x;
		double	v;

		// vertical
		fh = dev_textheight("0");
		n = dy / (fh * 1.5);
		
		if	( (n - 1) > 0 )	{
			for ( i = 0; i <= n; i ++ )	{
				if	( i == 0 )
					v = vmin;
				else if ( i == n )
					v = vmax;
				else
					v = vmin + ( ((double) i+1) * ((vmax-vmin)/ (double)(n+1)) );

				cmd_chart_fstr(v, buf);

				y = y1 + (dy - ((v-vmin)*ly));

				if	( i != 0 )	
					dev_setxy(rx1+1, y+1);
				else	
					dev_setxy(rx1+1, y-fh);

				dev_print(buf);
				dev_line(x1-4, y, x1, y);
				}
			}

		// horizontal
		fw = dev_textwidth("000");
		n = -1;
		if	( count <= 24 )	{
			if	( count * (fw * 1.34) < dx )
				n = count;
			}

		if	( n == -1 )
			n = dx / (fw * 1.5);
		
		if	( (n - 1) > 0 )	{
			for ( i = 0; i < n; i ++ )	{
				if	( i == 0 )
					v = 0;
				else
					v = i * ((double) count/ (double)n);

				ftostr(v+var_p->lbound[0], buf);
				buf[3] = '\0';
				fw = dev_textwidth(buf);

				x = x1 + v*lx;

				if	( chart == 1 )
					dev_setxy(x-fw, y2+1);
				else	{
					if ( x+fw+1 < x2 )
						dev_setxy(x+1, y2+1);
					}

				dev_print(buf);
				dev_line(x, y2, x, y2+4);
				}
			}

		dev_line(x1, y1, x1, y2);
		dev_line(x1, y2, x2, y2);

		x1 ++;
		y2 --;
		dx = (x2-x1);
		dy = (y2-y1);
		}

	// draw
	switch ( chart )	{

	case	1:	// line chart
		// draw line
		for ( i = 1; i < count; i ++ )	
			dev_line(pts[(i-1)*2], pts[(i-1)*2+1], pts[i*2], pts[i*2+1]);

		// draw marks
		if	( marks & 0x1 )	{
			for ( i = 0; i < count; i ++ )	{
				int		mx, my;
				int		fh, fw;

				cmd_chart_fstr(vals[i], buf);

				fw = dev_textwidth(buf);
				fh = dev_textheight(buf);

				mx = pts[i*2] - fw/2;
				my = pts[i*2+1];

				if	( my > (y1+(y2-y1)/2) )
					my -= fh;
				if	( mx <= x1 )
					mx = x1+1;
				if	( mx+fw >= x2 )
					mx = x2-fw;

				dev_setxy(mx, my);
				dev_print(buf);

				dev_rect(pts[i*2]-2, pts[i*2+1]-2, pts[i*2]+2, pts[i*2+1]+2, 1);
				}
			}
		break;

	case	2:	// bar chart
		// draw rect
		color = 0;
		for ( i = 1; i < count; i ++ )	{
			if	( os_color_depth > 2 )	{
				dev_setcolor(color);
				color ++;
				if	( color >= 15 )
					color = 0;
				}
			dev_rect(pts[(i-1)*2], pts[(i-1)*2+1], pts[i*2]-2, y2, 1);
			}

		if	( os_color_depth > 2 )	
			dev_setcolor(color);
		dev_rect(
			pts[(count-1)*2], pts[(count-1)*2+1], 
			pts[(count-1)*2]+lx-1, y2, 
			1);

		// draw marks
		if	( marks & 0x1 )	{
			color = 0;
			for ( i = 0; i < count; i ++ )	{
				int		mx, my;
				int		fh, fw;

				cmd_chart_fstr(vals[i], buf);

				fw = dev_textwidth(buf);
				fh = dev_textheight(buf);

				mx = pts[i*2] + lx/2-fw/2;
				my = pts[i*2+1];

				if	( os_color_depth > 2 )	{
					if	( my-fh >= y1 )	
						dev_settextcolor(0, 15);
					else	{
						if	( color >= 7 && color != 8 )
							dev_settextcolor(0, color);
						else
							dev_settextcolor(15, color);
						}

					color ++;
					if	( color >= 15 )
						color = 0;
					}

				if	( my-fh >= y1 )	
					my -= fh;

				if	( mx <= x1 )
					mx = x1 + 1;
				if	( mx+fw >= x2 )
					mx = x2 - fw;

				dev_setxy(mx, my);
				dev_print(buf);
				}
			}
		break;
		};

	//
	tmp_free(pts);
	tmp_free(vals);

	dev_settextcolor(prev_fgcolor, prev_bgcolor);
}
